# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import time
from StringIO import StringIO
from subprocess import Popen, PIPE
import textwrap
import xml.etree.ElementTree as ET
import re
import urllib2

from quixote import redirect, get_publisher
from quixote.directory import Directory
from quixote.html import TemplateIO, htmltext

from qommon import errors
from qommon import misc
from qommon.form import *
from qommon.backoffice.menu import html_top
from qommon.admin.menu import command_icon
from qommon import get_logger

from wcs.workflows import *
from wcs.formdef import FormDef
from wcs.formdata import Evolution
from .fields import FieldDefPage, FieldsDirectory


def svg(tag):
    return '{http://www.w3.org/2000/svg}%s' % tag

def xlink(tag):
    return '{http://www.w3.org/1999/xlink}%s' % tag

TITLE = svg('title')
POLYGON = svg('polygon')
XLINK_TITLE = xlink('title')

def remove_tag(node, tag):
    for child in node:
        if child.tag == tag:
            node.remove(child)

def remove_attribute(node, att):
    if att in node.attrib:
        del node.attrib[att]


def remove_style(node, top, colours, white_text=False):
    remove_tag(node, TITLE)
    if node.get('fill') == 'white' and node.get('stroke') == 'white':
        # this is the general white background, wipe it to be transparent
        node.attrib['fill'] = 'transparent'
        node.attrib['stroke'] = 'transparent'
    if node.tag == svg('text') and white_text:
        node.attrib['fill'] = 'white'
    for child in node:
        remove_attribute(child, XLINK_TITLE)
        if child.get('fill') in colours:
            matching_hexa = colours.get(child.get('fill'))
            child.attrib['fill'] = '#' + matching_hexa
            del child.attrib['stroke']
            if misc.get_foreground_colour(matching_hexa) == 'white':
                white_text = True
        if child.get('font-family'):
            del child.attrib['font-family']
        if child.get('font-size'):
            child.attrib['font-size'] = str(float(child.attrib['font-size'])*0.8)
        remove_attribute(child, 'style')
        remove_style(child, top, colours, white_text=white_text)

def graphviz_post_treatment(content, colours):
    ''' Remove all svg:title and top-level svg:polygon nodes, remove style
        attributes and xlink:title attributes.

        If a color style is set to a name matching class-\w+, set the second
        part on as class selector on the top level svg:g element.
    '''
    tree = ET.fromstring(content)
    style = ET.SubElement(tree, svg('style'))
    style.attrib['type'] = 'text/css'
    css_url = '%s%s%s' % (get_publisher().get_root_url(),
            get_publisher().qommon_static_dir,
            get_publisher().qommon_admin_css)
    style.text = '@import url(%s);' % css_url

    for root in tree:
        remove_tag(root, TITLE)
        # remove_tag(root, POLYGON)
        for child in root:
            remove_style(child, child, colours)
    return ET.tostring(tree)

def graphviz(workflow, url_prefix='', select=None, svg=True,
        include=False):
    out = StringIO()
    # a list of colours known to graphviz, they will serve as key to get back
    # to the colours defined in wcs.
    graphviz_colours = [
        'aliceblue', 'antiquewhite', 'aqua', 'aquamarine', 'azure', 'beige',
        'bisque', 'black', 'blanchedalmond', 'blue', 'blueviolet', 'brown',
        'burlywood', 'cadetblue', 'chartreuse', 'chocolate', 'coral',
        'cornflowerblue', 'cornsilk', 'crimson', 'cyan', 'darkblue',
        'darkcyan', 'darkgoldenrod', 'darkgray', 'darkgreen', 'darkgrey',
        'darkkhaki', 'darkmagenta', 'darkolivegreen', 'darkorange',
        'darkorchid', 'darkred', 'darksalmon', 'darkseagreen', 'darkslateblue',
        'darkslategray', 'darkslategrey', 'darkturquoise', 'darkviolet',
        'deeppink', 'deepskyblue', 'dimgray', 'dimgrey', 'dodgerblue',
        'firebrick', 'floralwhite', 'forestgreen', 'fuchsia', 'gainsboro',
        'ghostwhite', 'gold', 'goldenrod', 'gray', 'grey', 'green',
        'greenyellow', 'honeydew', 'hotpink', 'indianred', 'indigo', 'ivory',
        'khaki', 'lavender', 'lavenderblush', 'lawngreen', 'lemonchiffon',
        'lightblue', 'lightcoral', 'lightcyan', 'lightgoldenrodyellow',
        'lightgray', 'lightgreen', 'lightgrey', 'lightpink', ]

    colours = {}
    revert_colours = {}
    print >>out, 'digraph main {'
    # print >>out, 'graph [ rankdir=LR ];'
    print >>out, 'node [shape=box,style=filled];'
    print >>out, 'edge [];'
    for status in workflow.possible_status:
        i = status.id
        print >>out, 'status%s' % i,
        print >>out, '[label="%s"' % status.name,
        if select == str(i):
            print >>out, ',id=current_status'
        if status.colour:
            if status.colour not in colours:
                colours[status.colour] = graphviz_colours.pop()
                revert_colours[colours[status.colour]] = status.colour
            print >>out, ',color=%s' % colours[status.colour]
        print >>out, ' URL="%sstatus/%s/"];' % (url_prefix, i)

    for status in workflow.possible_status:
        i = status.id
        for item in status.items:
            next_status_ids = [x.id for x in item.get_target_status() if x.id != status.id]
            if not next_status_ids:
                next_status_ids = [status.id]
            for next_id in next_status_ids:
                print >>out, 'status%s -> status%s' % (i, next_id)
            url = 'status/%s/items/%s/' % (i, item.id)
            if getattr(item, 'label', None):
                label = item.label
                if getattr(item, 'by', None):
                    roles = workflow.render_list_of_roles(item.by)
                    label += ' %s %s' % (_('by'), roles)
            else:
                label = item.render_as_line()
            label = label.replace('"', '\\"')
            label = label.decode('utf8')
            label = textwrap.fill(label, 20, break_long_words=False)
            label = label.encode('utf8')
            label = label.replace('\n', '\\n')
            print >>out, '[label="%s"' % label,
            print >>out, ',URL="%s%s"]' % (url_prefix, url)

    print >>out, '}'
    out = out.getvalue()
    if svg:
        try:
            process = Popen(['dot', '-Tsvg'], stdin=PIPE, stdout=PIPE)
            out, err = process.communicate(out)
        except OSError:
            return ''
        out = graphviz_post_treatment(out, revert_colours)
        if include:
            # It seems webkit refuse to accept SVG when using its proper namespace,
            # and xlink namespace prefix must be xlink: to be acceptable
            out = out.replace('ns0:', '')
            out = out.replace('xmlns:ns0', 'xmlns:svg')
            out = out.replace('ns1:', 'xlink:')
            out = out.replace(':ns1', ':xlink')
    out = out.replace('<title>main</title>', '<title>%s</title>' % workflow.name)
    return out


class WorkflowUI(object):
    def __init__(self, workflow):
        self.workflow = workflow

    def form_new(self):
        form = Form(enctype='multipart/form-data')
        form.add(StringWidget, 'name', title = _('Workflow Name'), required = True, size=30)
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        return form

    def form_edit(self):
        form = Form(enctype='multipart/form-data')
        form.add_hidden('id', value = self.workflow.id)
        form.add(StringWidget, 'name', title = _('Workflow Name'), required = True, size=30,
                value = self.workflow.name)
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        return form

    def submit_form(self, form):
        if self.workflow:
            workflow = self.workflow
        else:
            workflow = Workflow(name = form.get_widget('name').parse())

        name = form.get_widget('name').parse()
        workflows_name = [x.name for x in Workflow.select() if x.id != workflow.id]
        if name in workflows_name:
            form.get_widget('name').set_error(_('This name is already used'))
            raise ValueError()

        for f in ('name',):
            setattr(workflow, f, form.get_widget(f).parse())
        workflow.store()
        return workflow

class WorkflowItemPage(Directory):
    _q_exports = ['', 'delete']

    def __init__(self, workflow, status, component, html_top):
        try:
            self.item = [x for x in status.items if x.id == component][0]
        except (IndexError, ValueError):
            raise errors.TraversalError()
        self.workflow = workflow
        self.status = status
        self.html_top = html_top
        get_response().breadcrumb.append(('items/%s/' % component, _(self.item.description)))

    def _q_index(self):
        request = get_request()
        if request.get_method() == 'GET' and request.form.get('file'):
            value = getattr(self.item, request.form.get('file'), None)
            if value:
                return value.build_response()

        form = Form(enctype='multipart/form-data')
        self.item.fill_admin_form(form)

        if not str(self.workflow.id).startswith(str('_')):
            form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_widget('cancel').parse():
            return redirect('..')

        if not form.get_submit() == 'submit' or form.has_errors():
            self.html_top('%s - %s' % (_('Workflow'), self.workflow.name))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s - %s</h2>') % (self.workflow.name, self.status.name)
            r += htmltext('<h3>%s</h3>') % _(self.item.description)
            r += form.render()
            if self.item.support_substitution_variables:
                r += htmltext('<h3>%s</h3>') % _('Substitution Variables')
                r += get_publisher().substitutions.get_substitution_html_table()
            return r.getvalue()
        else:
            self.item.submit_admin_form(form)
            self.workflow.store()
            return redirect('..')

    def delete(self):
        form = Form(enctype='multipart/form-data')
        form.widgets.append(HtmlWidget('<p>%s</p>' % _('You are about to remove an item.')))
        form.add_submit('delete', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        if form.get_widget('cancel').parse():
            return redirect('../../')
        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('delete', _('Delete')))
            self.html_top(title = _('Delete Item'))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s</h2>') % _('Deleting Item')
            r += form.render()
            return r.getvalue()
        else:
            del self.status.items[self.status.items.index(self.item)]
            self.workflow.store()
            return redirect('../../')

    def _q_lookup(self, component):
        t = self.item.q_admin_lookup(self.workflow, self.status, component,
                self.html_top)
        if t:
            return t
        return Directory._q_lookup(self, component)


class WorkflowItemsDir(Directory):
    _q_exports = ['']

    def __init__(self, workflow, status, html_top):
        self.workflow = workflow
        self.status = status
        self.html_top = html_top

    def _q_lookup(self, component):
        return WorkflowItemPage(self.workflow, self.status, component,
                self.html_top)

    def _q_index(self):
        return redirect('..')

class WorkflowStatusPage(Directory):
    _q_exports = ['', 'delete', 'newitem', ('items', 'items_dir'),
                  'update_order', 'edit', 'reassign', 'visibility',
                  'endpoint', 'colour']

    def __init__(self, workflow, status_id, html_top):
        self.html_top = html_top
        self.workflow = workflow
        try:
            self.status = [x for x in self.workflow.possible_status if x.id == status_id][0]
        except IndexError:
            raise errors.TraversalError()

        self.items_dir = WorkflowItemsDir(workflow, self.status, html_top)
        get_response().breadcrumb.append(('status/%s/' % status_id, _('Status: %s') % self.status.name))

    def _q_index(self):
        self.html_top('%s - %s' % (_('Workflow'), self.workflow.name))
        r = TemplateIO(html=True)
        get_response().add_javascript(['jquery.js', 'jquery-ui.js', 'biglist.js', 'svg-pan-zoom.js'])
        get_response().add_javascript_code('$(function () { svgPanZoom("svg", {controlIconsEnabled: true}); });');

        r += htmltext('<h2>%s - ') % _('Workflow')
        r += htmltext('%s - %s</h2>') % (self.workflow.name, self.status.name)
        r += get_session().display_message()

        r += htmltext('<div class="bo-block">')
        r += htmltext('<h3>%s ') % _('Possible Status:')
        r += htmltext('%s</h3>') % self.status.name
        r += htmltext('</div>')

        if self.status.visibility == ['_receiver']:
            r += htmltext('<div class="bo-block">')
            r += _('This status is hidden from the user.')
            if not str(self.workflow.id).startswith(str('_')):
                r += ' '
                r += htmltext('(<a href="visibility" rel="popup">%s</a>)') % _('change')
            r += htmltext('</div>')

        r += htmltext('<div class="bo-block">')
        if not self.status.items:
            r += htmltext('<p>%s</p>') % _('There are not yet any items in this status.')
        else:
            if str(self.workflow.id).startswith(str('_')):
                r += htmltext('<ul id="items-list" class="biglist">')
            else:
                r += htmltext('<p>')
                r += _('Use drag and drop to reorder items.')
                r += htmltext('</p>')
                r += htmltext('<ul id="items-list" class="biglist sortable">')
            for i, item in enumerate(self.status.items):
                r += htmltext('<li class="biglistitem" id="itemId_%s">') % item.id
                if str(self.workflow.id).startswith(str('_')):
                    r += item.render_as_line()
                else:
                    if hasattr(item, str('fill_admin_form')):
                        r += htmltext('<a href="items/%s/">%s</a>') % (item.id, item.render_as_line())
                    else:
                        r += item.render_as_line()
                    r += htmltext('<p class="commands">')
                    if hasattr(item, str('fill_admin_form')):
                        r += command_icon('items/%s/' % item.id, 'edit')
                    r += command_icon('items/%s/delete' % item.id, 'remove', popup = True)
                    r += htmltext('</p>')
                r += htmltext('</li>')
            r += htmltext('</ul>')
        r += htmltext('</div>') # bo-block

        source_status = []
        for status in self.workflow.possible_status:
            if status is self.status:
                continue
            for item in status.items:
                if self.status in item.get_target_status():
                    source_status.append(status)
                    break

        if source_status:
            r += htmltext('<div class="bo-block">')
            r += htmltext('<h3>%s</h3>') % _('Jumps')
            r += htmltext('<p>%s ') % _('This status is reachable from the following status:')
            r += htmltext(', ').join([htmltext('<a href="../%s/">%s</a>') % (x.id, x.name) for x in source_status])
            r += htmltext('.</p>')
            r += htmltext('</div>')

        r += htmltext('<p><a href="../../">%s</a></p>') % _('Back to workflow main page')

        r += htmltext('<div class="bo-block">')
        r += htmltext(graphviz(self.workflow, url_prefix='../../', include=True,
            select='%s' % self.status.id))
        r += htmltext('</div>')

        get_response().filter['sidebar'] = self.get_sidebar()

        return r.getvalue()


    def get_sidebar(self):
        get_response().add_javascript(['jquery.js', 'jquery-ui.js', 'popup.js',
            'jquery.colourpicker.js'])
        r = TemplateIO(html=True)
        if str(self.workflow.id).startswith(str('_')):
            r += htmltext('<p>')
            r += _('''This is the default workflow, you cannot edit it but you can
                 duplicate it to base your own workflow on it.''')
            r += htmltext('</p>')
        else:
            r += htmltext('<ul id="sidebar-actions">')
            r += htmltext('<li><a href="edit" rel="popup">%s</a></li>') % _('Change Status Name')
            r += htmltext('<li><a href="visibility" rel="popup">%s</a></li>') % _('Change Status Visibility')
            r += htmltext('<li><a href="endpoint" rel="popup">%s</a></li>') % _('Change Terminal Status')
            r += htmltext('<li><a href="colour" rel="popup">%s</a></li>') % _('Change Status Colour')
            r += htmltext('<li><a href="delete" rel="popup">%s</a></li>') % _('Delete')
            r += htmltext('</ul>')
            r += htmltext('<div id="new-field">')
            r += htmltext('<h3>%s</h3>') % _('New Item')
            r += self.get_new_item_form().render()
            r += htmltext('</div>')
        return r.getvalue()

    def get_new_item_form(self):
        form = Form(enctype='multipart/form-data', action = 'newitem')
        available_items = [x for x in item_classes if x.is_available()]
        def cmp_items(x, y):
            t = cmp(x.category and x.category[0], y.category and y.category[0])
            if t:
                return t
            return cmp(_(x.description), _(y.description))

        available_items.sort(cmp_items)
        options = [(x.key, _(x.description)) for x in available_items]
        form.add(SingleSelectWidget, 'type', title = _('Type'),
                required=True, options = options)
        form.add_submit('submit', _('Add'))
        return form

    def update_order(self):
        request = get_request()
        new_order = request.form['order'].strip(';').split(';')
        self.status.items = [ [x for x in self.status.items if x.id == y][0] for y in new_order]
        self.workflow.store()
        return 'ok'


    def newitem(self):
        form = self.get_new_item_form()

        if not form.is_submitted() or form.has_errors():
            get_session().message = ('error', _('Submitted form was not filled properly.'))
            return redirect('.')

        if form.get_widget('type').parse():
            self.status.append_item(form.get_widget('type').parse())
        else:
            get_session().message = ('error', _('Submitted form was not filled properly.'))
            return redirect('.')

        self.workflow.store()

        return redirect('.')


    def delete(self):
        form = Form(enctype="multipart/form-data")
        form.widgets.append(HtmlWidget('<p>%s</p>' % _(
                        "You are about to remove a status.")))
        form.add_submit('delete', _('Submit'))
        form.add_submit("cancel", _("Cancel"))
        if form.get_widget('cancel').parse():
            return redirect('../../')
        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('delete', _('Delete')))
            self.html_top(title = _('Delete Status'))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s %s</h2>') % (_('Deleting Status:'), self.status.name)
            r += form.render()
            return r.getvalue()
        else:
            # Before removing the status, scan formdata to know if it's in use.
            for formdef in FormDef.select():
                if formdef.workflow_id != self.workflow.id:
                    continue
                if formdef.data_class().get_with_indexed_value(
                                str('status'), 'wf-%s' % self.status.id):
                    return redirect('reassign')
            del self.workflow.possible_status[ self.workflow.possible_status.index(self.status) ]
            self.workflow.store()
            return redirect('../../')

    def edit(self):
        form = Form(enctype = 'multipart/form-data')
        form.add(StringWidget, 'name', title = _('Status Name'), required = True, size=30,
                value = self.status.name)
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        if form.get_widget('cancel').parse():
            return redirect('..')

        if form.is_submitted() and not form.has_errors():
            new_name = str(form.get_widget('name').parse())
            if [x for x in self.workflow.possible_status if x.name ==  new_name]:
                form.get_widget('name').set_error(
                                _('There is already a status with that name.'))
            else:
                self.status.name = new_name
                self.workflow.store()
                return redirect('.')

        self.html_top(title = _('Edit Workflow Status'))
        get_response().breadcrumb.append( ('edit', _('Edit')) )
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s</h2>') % _('Edit Workflow Status')
        r += form.render()
        return r.getvalue()

    def reassign(self):
        options = [(None, 'Do nothing'),
                    ('remove', _('Remove these forms'))]
        for status in self.workflow.get_waitpoint_status():
            if status.id == self.status.id:
                continue
            options.append(('reassign-%s' % status.id,
                _('Change these forms status to "%s"') % status.name))

        form = Form(enctype='multipart/form-data')
        form.add(SingleSelectWidget, 'action', title=_('Pick an Action'),
                 options=options)
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_submit() == 'cancel':
            return redirect('.')

        if form.is_submitted() and not form.get_widget('action').parse():
            return redirect('.')

        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('reassign', _('Delete / Reassign')))
            self.html_top(title = _('Delete Status / Reassign'))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s %s</h2>') % (_('Deleting Status:'), self.status.name)
            r += htmltext('<p>')
            r += _('''There are forms set to this status, they need to be changed before
                 this status can be deleted.''')
            r += htmltext('</p>')

            r += htmltext('<ul>')
            for formdef in FormDef.select():
                if formdef.workflow_id != self.workflow.id:
                    continue
                items = formdef.data_class().get_with_indexed_value(
                                str('status'), 'wf-%s' % self.status.id)
                r += htmltext('<li>%s: %s</li>') % (formdef.name, _('%s items') % len(items))
            r += htmltext('</ul>')

            r += form.render()
            return r.getvalue()
        else:
            self.submit_reassign(form)
            del self.workflow.possible_status[ self.workflow.possible_status.index(self.status) ]
            self.workflow.store()
            return redirect('../..')

    def submit_reassign(self, form):
        nb_forms = 0
        action = form.get_widget('action').parse()
        if action.startswith(str('reassign-')):
            new_status = 'wf-%s' % str(action)[9:]
        for formdef in FormDef.select():
            if formdef.workflow_id != self.workflow.id:
                continue
            for item in formdef.data_class().get_with_indexed_value(
                    str('status'), 'wf-%s' % self.status.id):
                nb_forms += 1
                if action == 'remove':
                    item.remove_self()
                else:
                    item.status = new_status
                    evo = Evolution()
                    evo.time = time.localtime()
                    evo.status = new_status
                    evo.comment = _('Administrator reassigned status')
                    if not item.evolution:
                        item.evolution = []
                    item.evolution.append(evo)
                    item.store()
            # delete all (old) status references in evolutions
            for item in formdef.data_class().select():
                if item.evolution:
                    modified = False
                    for evo in item.evolution:
                        if evo.status == self.status:
                            evo.status = None
                            modified = True
                    if modified:
                        item.store()
        if action == 'remove':
            get_logger().info('admin - delete status "%s" in workflow "%s": %d forms deleted' % (
                self.status.name, self.workflow.name, nb_forms))

    def visibility(self):
        form = Form(enctype = 'multipart/form-data')
        form.add(CheckboxWidget, 'hide_status_from_user',
                 title=_('Hide status from user'),
                 value=(self.status.visibility == ['_receiver']))
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        if form.get_widget('cancel').parse():
            return redirect('..')

        if form.is_submitted() and not form.has_errors():
            hide_status = form.get_widget('hide_status_from_user').parse()
            if hide_status:
                self.status.visibility = [str('_receiver')]
            else:
                self.status.visibility = None
            self.workflow.store()
            return redirect('.')

        self.html_top(title = _('Edit Workflow Status Visibility'))
        get_response().breadcrumb.append( ('visibility', _('Visibility')) )
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s</h2>') % _('Edit Workflow Status Visibility')
        r += form.render()
        return r.getvalue()

    def endpoint(self):
        form = Form(enctype = 'multipart/form-data')
        form.add(CheckboxWidget, 'force_terminal_status',
                 title=_('Force Terminal Status'),
                 value=(self.status.forced_endpoint == True))
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        if form.get_widget('cancel').parse():
            return redirect('..')

        if form.is_submitted() and not form.has_errors():
            self.status.forced_endpoint = form.get_widget('force_terminal_status').parse()
            self.workflow.store()
            return redirect('.')

        self.html_top(title = _('Edit Terminal Status'))
        get_response().breadcrumb.append( ('endpoint', _('Terminal Status')) )
        return form.render()

    def colour(self):
        form = Form(enctype = 'multipart/form-data')
        form.add(ColourWidget, 'colour',
                 title=_('Colour'),
                 value=self.status.colour)
        form.add_submit('submit', _('Submit'))
        form.add_submit('cancel', _('Cancel'))
        if form.get_widget('cancel').parse():
            return redirect('..')

        if form.is_submitted() and not form.has_errors():
            self.status.colour = form.get_widget('colour').parse()
            self.workflow.store()
            return redirect('.')

        self.html_top(title = _('Edit Status Colour'))
        get_response().breadcrumb.append( ('colour', _('Status Colour')) )
        return form.render()


class WorkflowStatusDirectory(Directory):
    _q_exports = ['']

    def __init__(self, workflow, html_top):
        self.workflow = workflow
        self.html_top = html_top

    def _q_lookup(self, component):
        return WorkflowStatusPage(self.workflow, component, self.html_top)

    def _q_index(self):
        return redirect('..')


class WorkflowVariableWidget(CompositeWidget):
    def __init__(self, name, value=None, workflow=None, **kwargs):
        CompositeWidget.__init__(self, name, **kwargs)
        if value and '*' in value:
            varname = None
        else:
            varname = value
        self.add(VarnameWidget, 'name', render_br=False, value=varname)
        options = []
        if workflow:
            for status in workflow.possible_status:
                for item in status.items:
                    prefix = '%s*%s*' % (status.id, item.id)
                    parameters = [x for x in item.get_parameters() if not getattr(item, x)]
                    label = getattr(item, 'label', None) or _(item.description)
                    for parameter in parameters:
                        key = prefix + parameter
                        fake_form = Form()
                        item.add_parameters_widgets(fake_form, [parameter])
                        parameter_label = fake_form.widgets[0].title
                        option_value = '%s / %s / %s' % (status.name, label, parameter_label)
                        options.append((key, option_value, key))
        if not options:
            return
        options = [('', '---', '')] + options
        self.widgets.append(HtmlWidget(
            _('or you can use this field to directly replace a workflow parameter:')))
        self.add(SingleSelectWidget, 'select', options=options,
                value=value,
                hint=_('This takes priority over a variable name'))

    def _parse(self, request):
        super(WorkflowVariableWidget, self)._parse(request)
        if self.get('select'):
            self.value = self.get('select')
        elif self.get('name'):
            self.value = self.get('name')


class WorkflowVariablesFieldDefPage(FieldDefPage):
    section = 'workflows'

    def form(self):
        form = super(WorkflowVariablesFieldDefPage, self).form()
        form.remove('varname')
        form.add(WorkflowVariableWidget, 'varname', title=_('Variable'),
                value=self.field.varname, advanced=False, required=True,
                workflow=self.objectdef.workflow)
        return form


class WorkflowVariablesFieldsDirectory(FieldsDirectory):
    _q_exports = ['', 'update_order', 'new']

    section = 'settings'
    field_def_page_class = WorkflowVariablesFieldDefPage
    support_import = False
    blacklisted_types = ['page']

    def index_top(self):
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s - %s - %s</h2>') % (_('Workflow'),
                self.objectdef.name, _('Variables'))
        r += get_session().display_message()
        if not self.objectdef.fields:
            r += htmltext('<p>%s</p>') % _('There are not yet any variables.')
        return r.getvalue()

    def index_bottom(self):
        pass


class VariablesDirectory(Directory):
    _q_exports = ['', 'fields']

    def __init__(self, workflow):
        self.workflow = workflow

    def _q_index(self):
        return redirect('fields/')

    def _q_traverse(self, path):
        get_response().breadcrumb.append(('variables/', _('Variables')))
        self.fields = WorkflowVariablesFieldsDirectory(
                WorkflowVariablesFieldsFormDef(self.workflow))
        return Directory._q_traverse(self, path)


class WorkflowPage(Directory):
    _q_exports = ['', 'edit', 'delete', 'newstatus', ('status', 'status_dir'), 'update_order',
            'duplicate', 'export', 'svg', ('variables', 'variables_dir')]

    def __init__(self, component, html_top):
        try:
            self.workflow = Workflow.get(component)
        except KeyError:
            raise errors.TraversalError()
        self.html_top = html_top
        self.workflow_ui = WorkflowUI(self.workflow)
        self.status_dir = WorkflowStatusDirectory(self.workflow, html_top)
        self.variables_dir = VariablesDirectory(self.workflow)
        get_response().breadcrumb.append((component + '/', self.workflow.name))

    def _q_index(self):
        self.html_top(title = _('Workflow - %s') % self.workflow.name)
        r = TemplateIO(html=True)
        get_response().add_javascript(['jquery.js', 'jquery-ui.js', 'biglist.js', 'svg-pan-zoom.js'])
        get_response().add_javascript_code('$(function () { svgPanZoom("svg", {controlIconsEnabled: true}); });');
        r += htmltext('<h2>%s - ') % _('Workflow')
        r += htmltext('%s') % self.workflow.name
        r += htmltext(' <span class="change">(<a rel="popup" href="edit">%s</a>)</span>') % _('change title')
        r += htmltext('</h2>')

        if self.workflow.last_modification_time:
            warning_class = ''
            if (time.time() - time.mktime(self.workflow.last_modification_time)) < 600:
                if get_request().user and str(get_request().user.id) != self.workflow.last_modification_user_id:
                    warning_class = 'recent'
            r += htmltext('<p class="last-modification %s">') % warning_class
            r += _('Last Modification:')
            r += ' '
            r += misc.localstrftime(self.workflow.last_modification_time)
            r += ' '
            if self.workflow.last_modification_user_id:
                try:
                    r += _('by %s') % get_publisher().user_class.get(
                                    self.workflow.last_modification_user_id).display_name
                except KeyError:
                    pass
            r += htmltext('</p>')

        r += get_session().display_message()

        r += htmltext('<div class="splitcontent-left">')

        r += htmltext('<div class="bo-block">')
        r += htmltext('<h3>%s</h3>') % _('Possible Status')

        if not self.workflow.possible_status:
            r += htmltext('<p>%s</p>') % _('There are not yet any status defined in this workflow.')
        else:
            if not str(self.workflow.id).startswith(str('_')):
                r += htmltext('<p>')
                r += _('Use drag and drop to reorder status.')
                r += htmltext('</p>')
                r += htmltext('<ul id="status-list" class="biglist sortable">')
            else:
                r += htmltext('<ul id="status-list" class="biglist">')
            for status in self.workflow.possible_status:
                klass = ['biglistitem']
                if status.visibility == ['_receiver']:
                    klass.append('hidden-status')
                r += htmltext('<li class="%s" id="itemId_%s">') % (
                        ' '.join(klass), status.id)
                r += htmltext('<a href="status/%s/">%s</a>') % (status.id, status.name)
                r += htmltext('</li>')
            r += htmltext('</ul>')
        r += htmltext('</div>')

        r += htmltext('</div>') # .splitcontent-left

        r += htmltext('<div class="splitcontent-right">')
        r += htmltext('<div class="bo-block">')
        r += htmltext('<h3>%s</h3>') % _('Workflow Roles')
        r += htmltext('<ul id="roles-list" class="biglist">')
        for key, label in (self.workflow.roles or {}).items():
            r += htmltext('<li class="bliglistitem">')
            r += htmltext('<a href="#roles/%s">%s</a>') % (key, label)
            r += htmltext('</li>')
        r += htmltext('</ul>')
        r += htmltext('</div>')

        if not str(self.workflow.id).startswith('_'):
            r += htmltext('<div class="bo-block">')
            r += htmltext('<h3>%s') % _('Workflow Variables')
            r += htmltext(' <span class="change">(<a href="variables/">%s</a>)</span></h3>') % _('change')
            if self.workflow.variables_formdef:
                r += htmltext('<ul class="biglist">')
                for field in self.workflow.variables_formdef.fields:
                    if field.varname:
                        r += htmltext('<li><a href="variables/fields/%s/">%s') % (
                                field.id, field.label)
                        if not '*' in field.varname:
                            r += htmltext(' (<code>%s</code>)') % field.varname
                        r += htmltext('</li>')
                r += htmltext('</ul>')
            r += htmltext('</div>')

        r += htmltext('</div>') # .splitcontent-right

        r += htmltext('<br style="clear:both;"/>')

        if self.workflow.possible_status:
            r += htmltext('<div class="bo-block">')
            r += htmltext(graphviz(self.workflow, include=True))
            r += htmltext('<div class="full-screen-link"><a href="svg">%s</a></div>') % _('Full Screen')
            r += htmltext('</div>') # bo-block

        formdefs = [x for x in FormDef.select() if x.workflow_id == self.workflow.id]
        if formdefs:
            r += htmltext('<div class="bo-block">')
            r += htmltext('<h3>%s</h3>') % _('Forms')
            r += htmltext('<p>%s ') % _('This workflow is used for the following forms:')
            r += htmltext(', ').join([htmltext('<a href="../../forms/%s/">%s</a>') % (x.id, x.name) for x in formdefs])
            r += htmltext('</p>')
            r += htmltext('</div>')

        get_response().filter['sidebar'] = self.get_sidebar()

        return r.getvalue()

    def get_sidebar(self):
        r = TemplateIO(html=True)
        if str(self.workflow.id).startswith(str('_')):
            r += htmltext('<p>')
            r += _('''This is the default workflow, you cannot edit it but you can
                 duplicate it to base your own workflow on it.''')
            r += htmltext('</p>')

        r += htmltext('<ul id="sidebar-actions">')
        if not str(self.workflow.id).startswith(str('_')):
            r += htmltext('<li><a href="delete" rel="popup">%s</a></li>') % _('Delete')
        r += htmltext('<li><a href="duplicate">%s</a></li>') % _('Duplicate')
        r += htmltext('<li><a href="export">%s</a></li>') % _('Export')
        r += htmltext('</ul>')
        if not str(self.workflow.id).startswith(str('_')):
            r += self.get_new_status_form()
        return r.getvalue()

    def svg(self):
        response = get_response()
        response.set_content_type('image/svg+xml')
        root_url = get_publisher().get_application_static_files_root_url()
        css = root_url + get_publisher().qommon_static_dir + get_publisher().qommon_admin_css
        return graphviz(self.workflow, include=False).replace('?>',
                '?>\n<?xml-stylesheet href="%s" type="text/css"?>\n' % css)

    def export(self):
        x = self.workflow.export_to_xml(include_id=True)
        misc.indent_xml(x)
        response = get_response()
        response.set_content_type('application/x-wcs-form')
        response.set_header('content-disposition',
                'attachment; filename=%s.wcs' % misc.simplify(self.workflow.name))
        return '<?xml version="1.0" encoding="utf-8"?>\n' + ET.tostring(x)

    def get_new_status_form(self):
        r = TemplateIO(html=True)
        r += htmltext('<div id="new-field">')
        r += htmltext('<h3>%s</h3>') % _('New Status')
        form = Form(enctype='multipart/form-data', action = 'newstatus')
        form.add(StringWidget, 'name', title = _('Name'), required = True, size = 50)
        form.add_submit('submit', _('Add'))
        r += form.render()
        r += htmltext('</div>')
        return r.getvalue()

    def update_order(self):
        request = get_request()
        new_order = request.form['order'].strip(';').split(';')
        self.workflow.possible_status = [ [x for x in self.workflow.possible_status if \
                                        x.id == y][0] for y in new_order]
        self.workflow.store()
        return 'ok'


    def newstatus(self):
        form = Form(enctype='multipart/form-data', action = 'newstatus')
        form.add(StringWidget, 'name', title = _('Name'), size = 50)

        if not form.is_submitted() or form.has_errors():
            get_session().message = ('error', _('Submitted form was not filled properly.'))
            return redirect('.')

        if form.get_widget('name').parse():
            try:
                self.workflow.add_status(form.get_widget('name').parse())
            except DuplicateStatusNameError:
                get_session().message = ('error', _('There is already a status with that name.'))
                return redirect('.')
        else:
            get_session().message = ('error', _('Submitted form was not filled properly.'))
            return redirect('.')

        self.workflow.store()

        return redirect('.')


    def edit(self, duplicate = False):
        form = self.workflow_ui.form_edit()
        if form.get_widget('cancel').parse():
            return redirect('..')

        if form.is_submitted() and not form.has_errors():
            try:
                self.workflow_ui.submit_form(form)
            except ValueError:
                pass
            else:
                return redirect('..')

        self.html_top(title = _('Edit Workflow'))
        r = TemplateIO(html=True)
        if duplicate:
            get_response().breadcrumb.append( ('edit', _('Duplicate')) )
            r += htmltext('<h2>%s</h2>') % _('Duplicate Workflow')
        else:
            get_response().breadcrumb.append( ('edit', _('Edit')) )
            r += htmltext('<h2>%s</h2>') % _('Edit Workflow')
        r += form.render()
        return r.getvalue()

    def delete(self):
        form = Form(enctype="multipart/form-data")
        for formdef in FormDef.select():
            if formdef.workflow_id == self.workflow.id:
                form.widgets.append(HtmlWidget('<p>%s</p>' % _(
                        "This workflow is currently in use, you cannot remove it.")))
                form.add_submit("cancel", _("Cancel"))
                break
        else:
            form.widgets.append(HtmlWidget('<p>%s</p>' % _(
                        "You are about to irrevocably delete this workflow.")))
            form.add_submit('delete', _('Submit'))
            form.add_submit("cancel", _("Cancel"))
        if form.get_widget('cancel').parse():
            return redirect('..')
        if not form.is_submitted() or form.has_errors():
            get_response().breadcrumb.append(('delete', _('Delete')))
            self.html_top(title = _('Delete Workflow'))
            r = TemplateIO(html=True)
            r += htmltext('<h2>%s %s</h2>') % (_('Deleting Workflow:'), self.workflow.name)
            r += form.render()
            return r.getvalue()
        else:
            self.workflow.remove_self()
            return redirect('..')

    def duplicate(self):
        self.workflow_ui.workflow.id = None
        original_name = self.workflow_ui.workflow.name
        self.workflow_ui.workflow.name = self.workflow_ui.workflow.name + _(' (copy)')
        workflow_names = [x.name for x in Workflow.select()]
        no = 2
        while self.workflow_ui.workflow.name in workflow_names:
            self.workflow_ui.workflow.name = _('%(name)s (copy %(no)d)') % {
                    'name': original_name, 'no': no}
            no += 1
        self.workflow_ui.workflow.store()
        return redirect('../%s/' % self.workflow_ui.workflow.id)


class WorkflowsDirectory(Directory):
    _q_exports = ['', 'new', ('import', 'p_import')]

    def html_top(self, title):
        return html_top('workflows', title)

    def _q_index(self):
        get_response().breadcrumb.append( ('workflows/', _('Workflows')) )
        self.html_top(title = _('Workflows'))
        r = TemplateIO(html=True)

        get_response().filter['sidebar'] = self.get_sidebar()

        r += htmltext('<div class="bo-block">')
        r += htmltext('<h2>%s</h2>') % _('Workflows')
        r += htmltext('<ul class="biglist">')
        for workflow in [Workflow.get_default_workflow()] + Workflow.select(order_by = 'name'):
            r += htmltext('<li>')
            r += htmltext('<strong class="label"><a href="%s/">%s</a></strong>') % (workflow.id, workflow.name)
            r += htmltext('</li>')
        r += htmltext('</ul>')
        r += htmltext('</div>') # .bo-block
        return r.getvalue()

    def get_sidebar(self):
        r = TemplateIO(html=True)
        r += htmltext('<ul id="sidebar-actions">')
        r += htmltext('  <li><a class="new-item" rel="popup" href="new">%s</a></li>') % _('New Workflow')
        r += htmltext('  <li><a href="import" rel="popup">%s</a></li>') % _('Import')
        r += htmltext('</ul>')
        return r.getvalue()

    def new(self):
        get_response().breadcrumb.append( ('workflows/', _('Workflows')) )
        get_response().breadcrumb.append( ('new', _('New')) )
        workflow_ui = WorkflowUI(None)

        form = workflow_ui.form_new()
        if form.get_widget('cancel').parse():
            return redirect('.')

        if form.is_submitted() and not form.has_errors():
            try:
                workflow = workflow_ui.submit_form(form)
            except ValueError:
                pass
            else:
                return redirect('%s/' % workflow.id)

        self.html_top(title = _('New Workflow'))
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s</h2>') % _('New Workflow')
        r += form.render()
        return r.getvalue()

    def _q_lookup(self, component):
        get_response().breadcrumb.append( ('workflows/', _('Workflows')) )
        return WorkflowPage(component, html_top=self.html_top)

    def p_import(self):
        form = Form(enctype = 'multipart/form-data')

        form.add(FileWidget, 'file', title=_('File'), required=False)
        form.add(UrlWidget, 'url', title=_('Address'), required=False,
                size=50)
        form.add_submit('submit', _('Import Workflow'))
        form.add_submit('cancel', _('Cancel'))

        if form.get_submit() == 'cancel':
            return redirect('.')

        if form.is_submitted() and not form.has_errors():
            try:
                return self.import_submit(form)
            except ValueError:
                pass

        get_response().breadcrumb.append( ('workflows/', _('Workflows')) )
        get_response().breadcrumb.append( ('import', _('Import')) )
        self.html_top(title=_('Import Workflow'))
        r = TemplateIO(html=True)
        r += htmltext('<h2>%s</h2>') % _('Import Workflow')
        r += htmltext('<p>%s</p>') % _(
                'You can install a new workflow by uploading a file '\
                'or by pointing to the workflow URL.')
        r += form.render()
        return r.getvalue()

    def import_submit(self, form):
        if form.get_widget('file').parse():
            fp = form.get_widget('file').parse().fp
        elif form.get_widget('url').parse():
            url = form.get_widget('url').parse()
            try:
                fp = urllib2.urlopen(url)
            except urllib2.HTTPError as e:
                form.set_error('url', _('Error loading form (%s).') % str(e))
                raise ValueError()
            except urllib2.URLError as e:
                form.set_error('url', _('Error loading form (%s).') % str(e))
                raise ValueError()
        else:
            form.set_error('file', _('You have to enter a file or a URL.'))
            raise ValueError()

        error, reason = False, None
        try:
            workflow = Workflow.import_from_xml(fp)
        except WorkflowImportError, e:
            error = True
            reason = _(e)
        except ValueError:
            error = True

        if error:
            if reason:
                msg = _('Invalid File (%s)') % reason
            else:
                msg = _('Invalid File')
            if form.get_widget('url').parse():
                form.set_error('url', msg)
            else:
                form.set_error('file', msg)
            raise ValueError()

        initial_workflow_name = workflow.name
        workflow_names = [x.name for x in Workflow.select()]
        copy_no = 1
        while workflow.name in workflow_names:
            if copy_no == 1:
                workflow.name = _('Copy of %s') % initial_workflow_name
            else:
                workflow.name = _('Copy of %(name)s (%(no)d)') % {
                        'name': initial_workflow_name, 'no': copy_no}
            copy_no += 1
        workflow.store()
        get_session().message = ('info',
                _('This workflow has been successfully imported.'))
        return redirect('%s/' % workflow.id)
